//@ts-check
"use strict";

/**
 * @typedef Notif
 * @type {{
 *   lifespan: number,
 *   node: PIXI.Text,
 * }}
 */

/**
@typedef Hud
@type {{
  pixiRoot: PIXI.Container;

  healthSprites: PIXI.TilingSprite[];
  itemSprites: PIXI.Sprite[];
  ammoEntries: { cartridge: PIXI.TilingSprite; icon: PIXI.Sprite; text: PIXI.Text; }[];
  ingredientEntries: { text: PIXI.Text; }[];

  craftContainer: PIXI.Container;
  recipesContainer: PIXI.Container;

  locationContainer: PIXI.Container;
  locationWorldText: PIXI.Text;
  locationTargetText: PIXI.Text;
  locationCountdown: number;

  notifsContainer: PIXI.Container;
  notifs: Notif[];

  isCrafting: boolean;
}}
 */

const hudMetrics = {
  bagPadLeft: 14,
  bagPadTop: 8,
  bagOutsideWidth: 226,
  bagOutsideHeight: 33,
}

/**
 * @returns Hud
 */
function createHud() {
  const pixiRoot = new PIXI.Container();
  inGame.pixiRoot.addChild(pixiRoot);

  // Health
  const healthSprites = [];

  for (let i = 0; i < inGame.stats.maxHealth + inGame.stats.maxArmor; i++) {
    const healthSprite = new PIXI.TilingSprite(tex("inGame/hud/health"), 30, 30);
    healthSprite.position.x = i * 30;
    healthSprite.tilePosition.x = i >= inGame.stats.maxHealth ? -60 : 0;
    pixiRoot.addChild(healthSprite);
    healthSprites.push(healthSprite);
  }

  // Items
  const bagSprite = new PIXI.Sprite(tex("inGame/hud/bagBg"));
  pixiRoot.addChild(bagSprite);
  bagSprite.anchor.x = 1;
  bagSprite.position.set(gameWidth, 0);

  const itemSprites = [];

  for (let i = 0; i < inGame.stats.itemSlots.length; i++) {
    const itemSprite = new PIXI.Sprite(tex("inGame/items/cheese"));
    itemSprite.position.set(gameWidth - 212 + i * 26, 3);
    itemSprite.visible = false;
    pixiRoot.addChild(itemSprite);
    itemSprites.push(itemSprite);
  }

  // Ammo
  const ammoEntries = [];

  const ammoSlotSize = 38;
  const ammoSlotSpacing = 4;
  const ammoStartX = gameWidth / 2 - (inGame.stats.ammoSlots.length * ammoSlotSize + (inGame.stats.ammoSlots.length - 1) * ammoSlotSpacing) / 2;
  for (let i = 0; i < inGame.stats.ammoSlots.length; i++) {
    const cartridge = new PIXI.TilingSprite(tex("inGame/hud/ammoSlot"), ammoSlotSize, ammoSlotSize);
    cartridge.position.set(ammoStartX + (ammoSlotSize + ammoSlotSpacing) * i, gameHeight - ammoSlotSize - ammoSlotSpacing);
    pixiRoot.addChild(cartridge);

    const icon = new PIXI.Sprite(tex("inGame/hud/ammo/pea"));
    icon.position.set(ammoStartX + (ammoSlotSize + ammoSlotSpacing) * i + ammoSlotSize / 2, gameHeight - ammoSlotSize - ammoSlotSpacing + ammoSlotSize / 2);
    icon.anchor.set(0.5, 0.5);
    pixiRoot.addChild(icon);

    const text = new PIXI.Text("x0", new PIXI.TextStyle({
      fontSize: 28, fontFamily: "Babyblocks",
      fill: "#ffddb2", strokeThickness: 4 * app.renderer.height / gameHeight, stroke: "#594449", lineJoin: "bevel"
    }));
    text.anchor.set(1, 1);
    text.scale.set(1 / 3);
    text.position.set(ammoStartX + (ammoSlotSize + ammoSlotSpacing) * i + ammoSlotSize - 2, gameHeight - ammoSlotSize - ammoSlotSpacing + ammoSlotSize - 2);
    pixiRoot.addChild(text);

    ammoEntries.push({ cartridge, icon, text });
  }

  // Ingredients
  const ingredientEntries = [];

  for (let i = 0; i < inGame.stats.ingredientSlots.length; i++) {
    const slot = inGame.stats.ingredientSlots[i];

    const icon = new PIXI.Sprite(tex(`inGame/hud/ingredients/${slot.plantDesc.id}`));
    icon.position.set(6 + 26 * i, 30);
    pixiRoot.addChild(icon);

    const text = new PIXI.Text("x0", new PIXI.TextStyle({
      fontSize: 28, fontFamily: "Babyblocks",
      fill: "#ffddb2", strokeThickness: 4 * app.renderer.height / gameHeight, stroke: "#594449", lineJoin: "bevel"
    }));
    text.scale.set(1 / 3);
    text.anchor.set(1, 0);
    text.position.set(30 + 26 * i, 38);
    pixiRoot.addChild(text);

    ingredientEntries.push({ text });
  }


  // Crafting
  const craftContainer = new PIXI.Container();
  pixiRoot.addChild(craftContainer);

  const recipesContainer = new PIXI.Container();
  craftContainer.addChild(recipesContainer);

  const ingredientsContainer = new PIXI.Container();
  pixiRoot.addChild(ingredientsContainer);

  // Notif
  const notifsContainer = new PIXI.Container();
  pixiRoot.addChild(notifsContainer);

  // Location
  const locationContainer = new PIXI.Container();
  locationContainer.position.set(gameWidth / 2, 40);
  locationContainer.visible = false;
  pixiRoot.addChild(locationContainer);
  const locationWorldText = new PIXI.Text("World", new PIXI.TextStyle({
    fontFamily: "Arvo", fontWeight: "bold", fontSize: 64,
    fill: "#ffddb2", strokeThickness: 12, stroke: "#594449", lineJoin: "bevel"
  }));
  locationWorldText.scale.set(1 / 3);
  locationWorldText.anchor.set(0.5, 0);
  locationContainer.addChild(locationWorldText);
  const locationTargetText = new PIXI.Text("Target", new PIXI.TextStyle({
    fontFamily: "Arvo", fontSize: 48,
    fill: "#ffddb2", strokeThickness: 12, stroke: "#594449", lineJoin: "bevel"
  }));
  locationTargetText.scale.set(1 / 3);
  locationTargetText.anchor.set(0.5, 0);
  locationTargetText.position.set(0, 32);
  locationContainer.addChild(locationTargetText);

  /** @type Hud */
  const hud = {
    pixiRoot,

    healthSprites,
    itemSprites,
    ammoEntries,
    ingredientEntries,

    craftContainer,
    recipesContainer,
    isCrafting: false,

    notifs: [],
    notifsContainer,

    locationContainer,
    locationWorldText,
    locationTargetText,
    locationCountdown: 0,
  };

  return hud;
}

/** @param {number} dt */
function updateHud(dt) {
  const { hud, stats } = inGame;

  // Crafting
  while (hud.recipesContainer.children.length > 0) hud.recipesContainer.children[0].destroy();

  if (hud.isCrafting) {
    let offsetX = (inGame.baxton.pos.x - inGame.camera.pos.x) * tileSize;
    let offsetY = (inGame.baxton.pos.y - inGame.camera.pos.y) * tileSize;

    let centerX = offsetX + gameWidth * .5;
    let centerY = offsetY + gameHeight * .5;

    // recipes
    {
      let baseX = centerX;
      let baseY = centerY - gameHeight * .1;

      let numSlots = stats.recipeSlots.length;

      let upwardsAngle = -Math.PI / 2;
      let anglePerSlot = Math.PI * .25;
      let reserveAngle = anglePerSlot;
      let totalAngle = reserveAngle * (numSlots - 1);

      let angleStart = upwardsAngle - totalAngle * .5;
      let radius = 50;

      let aimReference = new PIXI.Point(
        inGame.baxton.pos.x + (baseX - centerX) * inverseTileSize,
        inGame.baxton.pos.y + (baseY - 10 - centerY) * inverseTileSize,
      );
      let angle = Math.atan2(
        inGame.target.pos.y - aimReference.y,
        inGame.target.pos.x - aimReference.x,
      );
      if (angle < angleStart || angle > Math.PI / 2) {
        angle = angleStart;
      } else if (angle > angleStart + totalAngle) {
        angle = angleStart + totalAngle;
      }

      let bestDist = +Infinity;
      let currentRecipe = -1;
      for (let i = 0; i < numSlots; i++) {
        let slotAngle = angleStart + anglePerSlot * i;
        let dist = angleDist(slotAngle, angle);
        if (dist < bestDist) {
          bestDist = dist;
          currentRecipe = i;
        }
      }
      let currentSlot = stats.recipeSlots[currentRecipe];

      if (hud.isCrafting && inGame.input.mouse.releases[0]) {
        craftRecipe(currentSlot);
      }

      let wheel = new PIXI.Sprite(tex("inGame/hud/craftWheel"));
      wheel.anchor.set(.5, .5);
      wheel.position.set(baseX, baseY);
      hud.recipesContainer.addChild(wheel);

      let pointerCon = new PIXI.Container();
      pointerCon.position.set(baseX, baseY);
      const bla = (Math.PI / (stats.recipeSlots.length - 1));
      pointerCon.rotation = Math.round(angle / bla) * bla + Math.PI / 2;
      hud.recipesContainer.addChild(pointerCon);

      let pointer = new PIXI.Sprite(tex("inGame/hud/craftPointer"));
      pointer.position.set(0, -22);
      pointer.anchor.set(.5, .5);
      pointerCon.addChild(pointer);

      for (let i = 0; i < numSlots; i++) {
        let angle = angleStart + anglePerSlot * i;
        let x = baseX + Math.cos(angle) * radius;
        let y = baseY + Math.sin(angle) * radius;

        const slot = stats.recipeSlots[i];
        const disabled = !canAffordRecipe(slot);

        const bg = new PIXI.TilingSprite(tex("inGame/hud/recipeSlot"), 44, 44);
        bg.tilePosition.x = 0;
        if (currentRecipe === i) bg.tilePosition.x = -bg.width * (inGame.input.mouse.buttons[0] ? 1 : 2);
        bg.position.set(x, y);
        bg.anchor.set(.5, .5);

        let icon = new PIXI.Sprite(tex(`inGame/hud/ammo/${slot.result.id}`));
        icon.position.set(x, y);
        icon.anchor.set(.5, .5);

        let colorMatrix;
        if (disabled) {
          bg.tilePosition.x = -bg.width * 3;
          colorMatrix = new PIXI.filters.ColorMatrixFilter();
          colorMatrix.saturate(-0.7, true);
          colorMatrix.hue(180, true);
          icon.filters = [colorMatrix];
        }

        hud.recipesContainer.addChild(bg);
        hud.recipesContainer.addChild(icon);

        let ingredientStartAngle = -3 * Math.PI / 4;
        let ingredientAngleInc = Math.PI / 5;
        let ingredientRadius = 20;
        for (let j = 0; j < slot.ingredients.length; j++) {
          let ingDesc = slot.ingredients[j];
          let ingredientAngle = ingredientStartAngle + ingredientAngleInc * j;
          let ingredientX = x + Math.cos(ingredientAngle) * ingredientRadius;
          let ingredientY = y + Math.sin(ingredientAngle) * ingredientRadius;

          let icon = new PIXI.Sprite(tex(`inGame/hud/ingredients/${ingDesc.id}`));
          icon.anchor.set(.5, .5);
          icon.position.set(ingredientX, ingredientY);
          hud.recipesContainer.addChild(icon);

          if (colorMatrix) {
            icon.filters = [colorMatrix];
          }
        }
      }
    }
  }

  // Notifs
  {
    let deaths = false;
    for (const notif of hud.notifs) {
      notif.lifespan -= dt;
      if (notif.lifespan < 0) {
        // @ts-ignore c'mon type checker get your stuff together
        notif.node.destroy({ children: true });
        deaths = true;
      } else if (notif.lifespan < 1) {
        notif.node.y -= 24 * dt;
        notif.node.alpha *= 0.97;
      }
    }

    if (deaths) {
      hud.notifs = hud.notifs.filter((notif) => notif.lifespan > 0);
    }
  }
}

function updateHudHealth() {
  for (let i = 0; i < inGame.stats.maxHealth; i++) {
    const isFull = i < inGame.stats.health;
    inGame.hud.healthSprites[i].tilePosition.x = isFull ? 0 : -inGame.hud.healthSprites[i].width;
  }

  for (let i = 0; i < inGame.stats.maxArmor; i++) {
    inGame.hud.healthSprites[inGame.stats.maxHealth + i].visible = i < inGame.stats.armor;
  }
}

function updateHudAmmo() {
  const { stats, hud } = inGame;

  for (let i = 0; i < stats.ammoSlots.length; i++) {
    const slot = stats.ammoSlots[i];
    const entry = hud.ammoEntries[i];

    entry.cartridge.tilePosition.x = i == stats.currentAmmo ? -entry.cartridge.width : 0;

    if (slot && slot.ammoDesc && slot.quantity > 0) {
      entry.icon.visible = true;
      entry.text.visible = true;

      entry.icon.texture = tex(`inGame/hud/ammo/${slot.ammoDesc.id}`);
      entry.text.text = `x${slot.quantity}`;
    } else {
      entry.icon.visible = false;
      entry.text.visible = false;
    }
  }
}

function updateHudIngredients() {
  const { stats, hud } = inGame;

  for (let i = 0; i < stats.ingredientSlots.length; i++) {
    const slot = stats.ingredientSlots[i];
    const entry = hud.ingredientEntries[i];

    entry.text.text = `x${slot.quantity}`;
  }
}

function updateHudItem(slotNumber) {
  const { item } = inGame.stats.itemSlots[slotNumber];
  const sprite = inGame.hud.itemSprites[slotNumber];
  sprite.visible = item != null;
  if (item != null) sprite.texture = tex(`inGame/items/${item}`);
}

function showWorldTitle(worldName, targetName) {
  const { hud } = inGame;

  const title = getText(`worlds.${worldName}.title`, null, "");
  const subtitle = getText(`worlds.${worldName}.targets.${targetName}`, null, "");

  hud.locationContainer.visible = true;
  hud.locationContainer.alpha = 1;

  hud.locationWorldText.text = title;

  hud.locationTargetText.text = subtitle;
  hud.locationCountdown = 3;
}

function animateWorldTitle(dt) {
  const { hud } = inGame;

  if (hud.locationCountdown > 0) {
    hud.locationCountdown -= dt;

    hud.locationContainer.alpha = Math.min(1, hud.locationCountdown);
    if (hud.locationCountdown <= 0) hud.locationContainer.visible = false;
  }
}

/** @param {RecipeDesc} recipe @returns {boolean} */
function canAffordRecipe(recipe) {
  for (const i of recipe.ingredients) {
    if (!isCarryingIngredient(i)) {
      return false;
    }
  }
  return true;
}

/** @param {RecipeDesc} recipe @returns {boolean} */
function craftRecipe(recipe) {
  const { stats } = inGame;

  for (let i of recipe.ingredients) {
    if (!isCarryingIngredient(i)) {
      const plantName = getText(`plants.${i.id}`);
      notify("error", `Missing ${plantName}`);
      getSound("crafting.cantCraft").play();
      return false;
    }
  }

  let destSlotIndex = findAmmoSlot(recipe.result);
  if (destSlotIndex === null) {
    notify("error", "No free ammo slot!");
    getSound("crafting.cantCraft").play();
    return false;
  }

  for (let i of recipe.ingredients) {
    let slot = stats.ingredientSlots.find((s) => s.plantDesc === i);
    slot.quantity -= 1;
  }
  updateHudIngredients();

  let name = getText(`ammos.${recipe.result.id}`);
  notify("info", `Crafted x${recipe.amount} ${name}`);
  getSound("crafting.crafted").play();

  if (!stats.ammoSlots[destSlotIndex]) {
    stats.ammoSlots[destSlotIndex] = {
      ammoDesc: recipe.result,
      quantity: 0,
    };
  }
  stats.ammoSlots[destSlotIndex].ammoDesc = recipe.result;
  stats.ammoSlots[destSlotIndex].quantity += recipe.amount;
  updateHudAmmo();
}

/** @param {AmmoDesc} ammoDesc @returns {number | null} */
function findAmmoSlot(ammoDesc) {
  const { stats } = inGame;

  let existingIndex = stats.ammoSlots.findIndex((slot) => slot && slot.ammoDesc === ammoDesc);
  if (existingIndex != -1) {
    return existingIndex;
  }

  let freeIndex = stats.ammoSlots.findIndex((slot) => slot === null);
  if (freeIndex != -1) {
    return freeIndex;
  }

  let zeroIndex = stats.ammoSlots.findIndex((slot) => slot && slot.quantity <= 0);
  if (zeroIndex != -1) {
    return zeroIndex;
  }

  return null;
}

/** @param {PlantDesc} ingredient @returns {boolean} */
function isCarryingIngredient(ingredient) {
  const { stats } = inGame;

  for (let slot of stats.ingredientSlots) {
    if (slot.plantDesc === ingredient && slot.quantity > 0) {
      return true;
    }
  }
  return false;
}

/**
 * @typedef NotificationType
 * @type {"info" | "error" | "huge"}
 */

/**
 * @param {NotificationType} type
 * @param {string} msg
 */
function notify(type, msg) {
  const { hud } = inGame;

  /** @type {string} */
  let color;
  /** @type {string} */
  let outline;
  switch (type) {
    case "error":
      color = "#cf3a30";
      outline = "#521d19";
      break;
    case "info":
    default:
      color = "#ffddb2";
      outline = "#594449";
      break;
  }

  /** @type {Partial<PIXI.TextStyle>} */
  let style = {
    fontSize: 32, fontFamily: "Babyblocks",
    fill: color, strokeThickness: app.renderer.height / gameHeight * 2, stroke: outline, lineJoin: "bevel"
  };

  if (type === "huge") {
    style.fontFamily = "Arvo";
    style.fontSize = 70;
    style.strokeThickness *= 2;
  }

  let txt = new PIXI.Text(msg, style);
  txt.anchor.set(.5, .5);
  txt.position.set(
    gameWidth * .5,
    gameHeight * .5 + 80,
  );
  txt.scale.set(1 / 3);
  hud.notifsContainer.addChild(txt);
  hud.notifs.push({
    lifespan: 1.5,
    node: txt,
  });
}
